# -*- coding: iso-8859-1 -*-
#
# Copyright (C) 2006 Herbert Valerio Riedel <hvr@gnu.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import os, re, random, shutil
#from traceback import print_stack

class GitError(Exception):
    pass

class Storage:
    def __init__(self,repo):
        self.repo = repo
        self.commit_encoding = None
        self.cache_dir = '/var/cache/trac-git/' + self.repo.replace ('/', '__')

        try:
            os.makedirs (self.cache_dir, 0755)
        except:
            pass

        try:
            os.makedirs (self.cache_dir + '/short', 0755)
        except:
            pass

        try:
            os.makedirs (self.cache_dir + '/long', 0755)
        except:
            pass

        current_head = self._head_uncached ()

        try:
            f = open (self.cache_dir + '/head', 'r')
        except:
            cached_head = "invalid"
        else:
            cached_head = f.read ()
            f.close ()

        if current_head != cached_head:
            try:
                shutil.rmtree (self.cache_dir + '/short', False)
            except:
                pass

            try:
                os.makedirs (self.cache_dir + '/short', 0755)
            except:
                pass

            f = open (self.cache_dir + '/head', 'w')
            f.write (current_head)
            f.close ()

    def _git_call_f(self,cmd,long_term_cache=False):
        #f=open('/tmp/gitplugin.log', 'a')
        #f.write(cmd + '\n')
        #f.close()

        # [HPJ] Add caching.

        if long_term_cache is True:
            cache_name = self.cache_dir + '/long/'
        else:
            cache_name = self.cache_dir + '/short/'

        cache_name = cache_name + cmd.encode ('base64_codec')

        if not os.path.isfile (cache_name):
            (input, output, error) = os.popen3 ('GIT_DIR="%s" %s' % (self.repo,cmd))
            temp_name = '/tmp/gitplugin-%s-%s' % (int (random.getrandbits (32)), os.getpid ())

            try:
                cache_file = open (temp_name, "w")
            except:
                return output

            blob = output.read ()
            output.close ()
            cache_file.write (blob)
            cache_file.close ()
            shutil.move (temp_name, cache_name)

        cache_file = open (cache_name, "r")
        return cache_file

    def _git_call(self,cmd,use_cache=False):
        return self._git_call_f(cmd, use_cache).read()

    def get_commit_encoding(self):
        if self.commit_encoding is None:
            self.commit_encoding = self._git_call("git-repo-config --get i18n.commitEncoding", True).strip()
            if ''==self.commit_encoding:
                self.commit_encoding = 'utf-8'
        return self.commit_encoding

    def _head_uncached(self):
        (input, output, error) = os.popen3 ('GIT_DIR="%s" git-rev-parse --verify HEAD' % (self.repo))
        head = output.read().strip()
        output.close()
        return head

    def head(self):
        "get current HEAD commit id"
        return self.verifyrev("HEAD")

    def verifyrev(self,rev):
        "verify/lookup given revision object and return a sha id or None if lookup failed"
        rc=self._git_call("git-rev-parse --verify '%s'" % rev, False).strip()
        if len(rc)==0:
            return None
        return rc

    def shortrev(self,rev):
        "try to shorten sha id"
        return self._git_call("git-rev-parse --short '%s'" % rev, False).strip()

    def get_branches(self):
        "returns list of branches, with active (i.e. HEAD) one being the first item"
        result=[]
        for e in self._git_call_f("git-branch", False).readlines():
            bname=e[1:].strip()
            if e[0]=='*':
                result.insert(0,bname)
            else:
                result.append(bname)
        return result

    def tree_ls(self,sha,path=""):
        if len(path)>0 and path[0]=='/':
            path=path[1:]
        return [e[:-1].split(None, 3) for e in self._git_call_f("git-ls-tree %s '%s'" % (sha,path), False).readlines()]

    def read_commit(self, sha):
        raw = self._git_call("git-cat-file commit "+sha, True)
        raw = unicode(raw, self.get_commit_encoding(), 'replace')
        lines = raw.splitlines()

        if len (lines) > 0:
            line = lines.pop(0)
        else:
            line = ""

        d = {}
        while line != "":
            (key,value)=line.split(None, 1)
            if not d.has_key(key):
                d[key] = []
            d[key].append(value.strip())
            line = lines.pop(0)

        if not d.has_key("committer"): 
            d["committer"] = [ "Unknown <nobody@example.com> 1187972517 -0800" ]  
        if not d.has_key("author"): 
            d["author"] = [ "Unknown <nobody@example.com> 1187972517 -0800" ] 

        return ("\n".join(lines),d)

    def get_file(self, sha):
        return self._git_call_f("git-cat-file blob "+sha, True)

    def get_obj_size(self, sha):
        return int(self._git_call("git-cat-file -s "+sha, True).strip())

    def parents(self, sha):
        tmp=self._git_call("git-rev-list --max-count=1 --parents "+sha, False)
        tmp=tmp.strip()
        tmp=tmp.split()
        return tmp[1:]

    def children(self, sha):
        for revs in self._git_call_f("git-rev-list --parents HEAD", False).readlines():
            revs = revs.strip()
            revs = revs.split()
            if sha in revs[1:]:
                yield revs[0]
        
    def history(self, sha, path, skip=0):
        for rev in self._git_call_f("git-rev-list %s -- '%s'" % (sha,path), False).readlines():
            if(skip > 0):
                skip = skip - 1
                continue
            yield rev.strip()

    def last_change(self, sha, path):
        for rev in self._git_call_f("git-rev-list --max-count=1 %s -- '%s'" % (sha,path), False).readlines():
            return rev.strip()
        return None

    def diff_tree(self, tree1, tree2, path=""):
        if tree1 is None:
            tree1 = "--root"
        cmd = "git-diff-tree -r %s %s -- '%s'" % (tree1, tree2, path)
        for chg in self._git_call_f(cmd, True).readlines():
            if chg.startswith(tree2):
                continue
            (mode1,mode2,obj1,obj2,action,path) = chg[:-1].split(None, 5)
            mode1 = mode1[1:]
            yield (mode1,mode2,obj1,obj2,action,path)

if __name__ == '__main__':
    import sys
    
    g = Storage(sys.argv[1])
    
    print "[%s]" % g.head()
    print g.tree_ls(g.head())
    print "--------------"
    print g.read_commit(g.head())
    print "--------------"
    print g.parents(g.head())

    print "--------------"
    print g.get_commit_encoding()
    print "--------------"
    print g.get_branches()
